/*
 * modified for lecture by Fabian Fuchs (KIT, Karlsruhe) based on:
 * 
 Copyright (c) 2007, Distributed Computing Group (DCG)
                    ETH Zurich
                    Switzerland
                    dcg.ethz.ch

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

 - Neither the name 'Sinalgo' nor the names of its contributors may be
   used to endorse or promote products derived from this software
   without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package projects.uebung2solution.nodes.nodeImplementations;


import java.awt.Color;
import java.awt.Graphics;
import java.util.LinkedList;

import javax.swing.JOptionPane;

import projects.defaultProject.nodes.timers.MessageTimer;
import projects.uebung2.nodes.messages.DefaultMessage;
import sinalgo.configuration.WrongConfigurationException;
import sinalgo.gui.helper.NodeSelectionHandler;
import sinalgo.gui.transformation.PositionTransformation;
import sinalgo.nodes.Node;
import sinalgo.nodes.Position;
import sinalgo.nodes.messages.Inbox;
import sinalgo.nodes.messages.Message;
import sinalgo.runtime.GUIRuntime;
import sinalgo.runtime.Main;
import sinalgo.tools.Tools;
import sinalgo.tools.Tuple;




/**
 * A leader election node 
 */
public class GreedyRoutingNode extends Node {

	/* store the current round */
	private int round;

	/* list of list of neighbors with their positions */
	private LinkedList<Tuple<Integer,Position>> neighborPositions;
	
	/* add additional variables / flags here */
	
	
	/* neighborhood learning payload*/
	public class NLPayload {
		public int fromID;
		public Position fromPosition;
		public boolean isResponse;
		public NLPayload (int id, Position pos, boolean response) {
			fromID = id;
			fromPosition = pos;
			isResponse = response;
		}
	}
	
	/* payload for message */
	public class GRPayload<T> {
		public Position toPosition;
		public T payload;
		public GRPayload (Position to, T pl) {
			toPosition = to;
			payload = pl;
		}
	}
	
	/* 
	 * finds the closest neighbor based on this.neighborPositions
	 * and return the ID of the closest neighbor (or the own ID)
	 */
	private int getNextNodeID(Position destination) {
		//sanity check:
		if (this.outgoingConnections.size() != this.neighborPositions.size()) {
			Tools.appendToOutput("ID" + this.ID + ": neighborPositions not initialized correctly: " + this.outgoingConnections.size() + " neighbors but " + this.neighborPositions.size() + " positions\n");
		}
		
		//init to the node itself
		double minDist = this.computeDistance(this.getPosition(), destination);
		int routeTo = this.ID;
		double tmp = 0;
		//try to find a neighbor with better values
		for (Tuple<Integer,Position> t : this.neighborPositions) {
			tmp = this.computeDistance(t.second, destination);
			if (tmp < minDist) {
				minDist = tmp;
				routeTo = t.first;
			}
		}
		//return best value
		return routeTo;
	}
	
	private double computeDistance(Position from, Position to) {
		double xd = Math.abs(from.xCoord - to.xCoord);
		double yd = Math.abs(from.yCoord - to.yCoord);
		double zd = Math.abs(from.zCoord - to.zCoord);
		return Math.sqrt(xd*xd+yd*yd+zd*zd);
	}
	
	@Override
	public void handleMessages(Inbox inbox) {
		while(inbox.hasNext()) {
			Message msg = inbox.next();
			if(msg instanceof DefaultMessage<?>) {
				if (((DefaultMessage<?>) msg).getLoad() instanceof NLPayload) {
					//received message, store information if not already stored 
					NLPayload pl = (NLPayload)((DefaultMessage<?>) msg).getLoad();
					Tuple<Integer,Position> information = new Tuple<Integer,Position>(pl.fromID,pl.fromPosition);
					if (!this.neighborPositions.contains(information)) {
						this.neighborPositions.add(information);	
					}
					//if the message is not a response itself, answer the request
					if (!pl.isResponse ){
						NLPayload response = new NLPayload(this.ID, this.getPosition(),true);
						this.send(new DefaultMessage<NLPayload>(response), Tools.getNodeByID(pl.fromID));
					}
				}
				
				
				if (((DefaultMessage<?>) msg).getLoad() instanceof GRPayload<?>) {
					//unpack packet
					GRPayload<?> packet = (GRPayload<?>)((DefaultMessage<?>) msg).getLoad();
					//I assume a String payload
					if (!(packet.payload instanceof String)){
						Tools.appendToOutput("Payload error. Payload: " + packet.payload);
					}
					//is packet for me?
					if (this.getPosition() == packet.toPosition) {
						Tools.appendToOutput("ID " + this.ID + ": Received message with payload " + (String)(packet.payload) + "\n");
					} else {
						//find closest neighbor
						int routeTo = getNextNodeID(packet.toPosition);
						//if closest neighbor is not the node itself, transmit the packet to the neighbor
						if (routeTo != this.ID) {
							Tools.appendToOutput("ID " + this.ID + ": Forward message\n");
							this.send(msg, Tools.getNodeByID(routeTo));
						} else {
							Tools.appendToOutput("ID " + this.ID + ": Message got stuck\n");
						}
					} 
				}
			}
		}
	}

	@Override
	public void preStep() {
		//executed before each 'step' (i.e. timers and handleMessages)
		//I would personally recommend to only use the postStep method for simplicity
	}
	
	@Override
	public void postStep() {
		//executed after each 'step' (i.e. timers and handleMessages)
		round++;
	}

	@Override
	public void init() {
		// initialize the node / read configuration values
		round = 0;
		this.neighborPositions = new LinkedList<Tuple<Integer,Position>>();
	}

	@Override
	public void neighborhoodChange() {
		//used / relevant only for initialization and mobility
		this.neighborPositions.clear();
		this.broadcast(new DefaultMessage<NLPayload>(new NLPayload(this.ID, this.getPosition(),false)));
	}
		
	/* (non-Javadoc)
	 * @see sinalgo.nodes.Node#draw(java.awt.Graphics, sinalgo.gui.transformation.PositionTransformation, boolean)
	 */
	public void draw(Graphics g, PositionTransformation pt, boolean highlight) {
		//set the color of this node, this allows us to see how the most important maxIDs spread
		
		this.setColor(Color.BLACK); 
		String text = Integer.toString(this.ID);
		// draw the node as a circle with the text inside
		super.drawNodeAsDiskWithText(g, pt, highlight, text, 10, Color.YELLOW);
	}


	@Override
	public String toString() {
		return "Node " + this.ID; 
	}

	@Override
	public void checkRequirements() throws WrongConfigurationException {
		//does configuration fit the nodes requirements? 
		
	}
	
	/*
	 * Starts the greedy routing of a message
	 */
	@NodePopupMethod(menuText="StartRouting")
	public void startRouting() {
		
		class NodeSelectionHandlerGreedyRouting implements NodeSelectionHandler {
			public GreedyRoutingNode sender;
			public void handleNodeSelectedEvent(Node n) { //n = null implies abort by user
				if (n == null) {
					Tools.appendToOutput("GreedyRouting: Abort by user\n");
					return;
				}
				sender.doStartRouting(n);
			}
			NodeSelectionHandlerGreedyRouting(GreedyRoutingNode sender) {
				this.sender = sender;
			}

		}
		NodeSelectionHandlerGreedyRouting nsh = new NodeSelectionHandlerGreedyRouting(this);

		Tools.getNodeSelectedByUser(nsh, "Select the destination node.");
	}
	
	public void doStartRouting(Node to) {
		if (round < 3) {
			JOptionPane.showMessageDialog(((GUIRuntime)Main.getRuntime()).getGUI(), "GreedyRoutingNodes gather their neighbors positions in the first two rounds, start routing afterwards.");
			return;
		}
		//create payload
		String payload = JOptionPane.showInputDialog(null, "Enter a payload String");

		Position destination = to.getPosition();
		DefaultMessage<GRPayload<String>> msg = new DefaultMessage<GRPayload<String>>(new GRPayload<String>(destination, payload));
		//compute the best neighbor 
		int routeTo = getNextNodeID(destination);
		//if closest neighbor is not the node itself, transmit the packet to the neighbor
		if (routeTo != this.ID) {
			MessageTimer msgTimer = new MessageTimer(msg, Tools.getNodeByID(routeTo));
			msgTimer.startRelative(1, this);
			Tools.appendToOutput("ID: " + this.ID + " Start Greedy Routing a packet to " + to.ID + " with position " + destination + "\n");		
		} else {
			Tools.appendToOutput("ID " + this.ID + " Could not send message, Node is closest neighbor");
		}
	}
}
